# ==============================================================================
# Medical Image Registration ToolKit (MIRTK)
#
# Copyright 2013-2016 Imperial College London
# Copyright 2013-2016 Andreas Schuh
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

##############################################################################
# @file  MIRTKRegistrationConfig.cmake
# @brief CMake package configuration file.
#
# @note The MIRTKRegistrationConfig.cmake file is automatically generated
#       by BASIS from the template file Config.cmake.in which is part of
#       the BASIS package.
#
# Variables in this file have to be prefixed by the name of this package.
# In the default package configuration file of BASIS, which can be included
# here using the \@BASIS_CONFIG\@ variable, this prefix is defined as @c NS
# variable. Moreover, the macro basis_set_config() should be used for
# convenience to set any configuration variable, e.g.,
#
# @code
# basis_set_config (INCLUDE_DIR "\@INCLUDE_DIR_CONFIG\@")
# @endcode
#
# where the variable @c INCLUDE_DIR_CONFIG is set in the ConfigSettings.cmake
# file of the BASIS installation. Note that this statement will set the
# variable ${NS}INCLUDE_DIR, where by default @c NS is the package name
# followed by an underscore (_).
#
# See the Config.cmake.in file of BASIS for details.
#
# @ingroup BasisConfig
##############################################################################

## @addtogroup BasisConfig
# @{


# ============================================================================
# namespace
# ============================================================================

# prefix used for variable names
set (NS "MIRTK_Registration_")

# allow caller to change namespace - used by projects with modules
if (${NS}CONFIG_PREFIX)
  set (NS "${${NS}CONFIG_PREFIX}_")
endif ()

# ----------------------------------------------------------------------------
macro (basis_set_config NAME)
  set (${NS}${NAME} ${ARGN})
endmacro ()

# ----------------------------------------------------------------------------
macro (basis_copy_config NAME SRC)
  set (${NS}${NAME} ${${NS}${SRC}})
endmacro ()

# ----------------------------------------------------------------------------
macro (basis_append_config NAME)
  list (APPEND ${NS}${NAME} ${ARGN})
endmacro ()

# ----------------------------------------------------------------------------
macro (basis_remove_duplicates_config NAME)
  if (${NS}${NAME})
    list (REMOVE_DUPLICATES ${NS}${NAME})
  endif ()
endmacro ()

# ----------------------------------------------------------------------------
macro (basis_remove_item_config NAME ITEM)
  if (${NS}${NAME})
    list (REMOVE_ITEM ${NS}${NAME} "${ITEM}")
  endif ()
endmacro ()

# ============================================================================
# general
# ============================================================================

## @brief Installation prefix or root of build tree.
basis_set_config (INSTALL_PREFIX "${CMAKE_CURRENT_LIST_DIR}/../../../")
get_filename_component (${NS}INSTALL_PREFIX "${${NS}INSTALL_PREFIX}" ABSOLUTE)

## @brief Project use file to be included by other projects.
basis_set_config (USE_FILE "${CMAKE_CURRENT_LIST_DIR}/MIRTKRegistrationUse.cmake")

## @brief Directory of example files.
basis_set_config (EXAMPLE_DIR "")

# ============================================================================
# libraries
# ============================================================================

## @brief Directory of header files.
basis_set_config (INCLUDE_DIR "${${NS}INSTALL_PREFIX}/include")

## @brief Directory of header files including dependencies.
basis_set_config    (INCLUDE_DIRS) # to have a set recognized by Doxygen filter
basis_copy_config   (INCLUDE_DIRS INCLUDE_DIR)
basis_append_config (INCLUDE_DIRS "")

## @brief Directory of libraries.
basis_set_config (LIBRARY_DIR "${${NS}INSTALL_PREFIX}/lib/mirtk")

## @brief Directory of libraries including dependencies.
basis_set_config    (LIBRARY_DIRS) # to have a set() recognized by Doxygen filter
basis_copy_config   (LIBRARY_DIRS LIBRARY_DIR)
basis_append_config (LIBRARY_DIRS "")

## @brief Search path for Python modules.
basis_set_config (PYTHONPATH "${${NS}INSTALL_PREFIX}/lib/mirtk/python")

## @brief Search path for Jython modules.
basis_set_config (JYTHONPATH "${${NS}INSTALL_PREFIX}/lib/mirtk/jython")

## @brief Search path for Perl modules.
basis_set_config (PERL5LIB "${${NS}INSTALL_PREFIX}/lib/mirtk/perl")

## @brief Search path for MATLAB modules.
basis_set_config (MATLABPATH "${${NS}INSTALL_PREFIX}/lib/mirtk/matlab")

## @brief Search path for Bash modules.
#
# See the import() function of the BASIS Bash utilities.
basis_set_config (BASHPATH "${${NS}INSTALL_PREFIX}/lib/mirtk/bash")

# ============================================================================
# exports
# ============================================================================

## @brief Absolute path of exports file to be included by other projects.
basis_set_config (EXPORTS_FILE "${CMAKE_CURRENT_LIST_DIR}/MIRTKRegistrationTargets.cmake")

## @brief Absolute path of custom exports file to be included by other projects.
basis_set_config (CUSTOM_EXPORTS_FILE "${CMAKE_CURRENT_LIST_DIR}/MIRTKRegistrationCustomTargets.cmake")

# ============================================================================
# dependencies
# ============================================================================

set (_depwarn " set to different value than during the configuration of Registration."
              " Using different versions of a dependency may cause inconsistencies!")
# VTK
if (DEPENDS_VTK_DIR)
  if (NOT DEPENDS_VTK_DIR STREQUAL "/usr/local/lib/cmake/vtk-9.2")
    message (WARNING DEPENDS_VTK_DIR ${_depwarn})
  endif ()
else ()
  basis_set_or_update_value (DEPENDS_VTK_DIR "/usr/local/lib/cmake/vtk-9.2")
endif ()
# TBB
if (DEPENDS_TBB_DIR)
  if (NOT DEPENDS_TBB_DIR STREQUAL "/usr/local")
    message (WARNING DEPENDS_TBB_DIR ${_depwarn})
  endif ()
else ()
  basis_set_or_update_value (DEPENDS_TBB_DIR "/usr/local")
endif ()
unset (_depwarn)


## @brief Whether this library was built with PointSet support
basis_set_config(WITH_PointSet FALSE)

## @brief Whether this library was built with Deformable support
basis_set_config(WITH_Deformable FALSE)


## @}
# end of Doxygen group - keep at end of file
